"""Custom Haystack Search Backend."""
from logging import getLogger
from multiprocessing import cpu_count

from django.utils.timezone import now
from haystack.backends.whoosh_backend import TEXT, WhooshSearchBackend
from haystack.constants import DJANGO_ID
from haystack.exceptions import SkipDocument
from humanfriendly import InvalidSize, parse_size
from whoosh.analysis import CharsetFilter, StandardAnalyzer, StemFilter
from whoosh.fields import NUMERIC
from whoosh.qparser import FieldAliasPlugin, GtLtPlugin, OperatorsPlugin
from whoosh.qparser.dateparse import DateParserPlugin
from whoosh.query import Not, Or, Term
from whoosh.support.charset import accent_map

from codex.librarian.search.status import SearchIndexStatusTypes
from codex.logger.logging import get_logger
from codex.search.indexes import ComicIndex
from codex.search.writing import CodexWriter
from codex.worker_base import WorkerBaseMixin


def gen_multipart_field_aliases(field):
    """Generate aliases for fields made of snake_case words."""
    bits = field.split("_")
    aliases = []

    # Singular from plural
    if field.endswith("s"):
        aliases += [field[:-1]]

    # Alternate delimiters
    for connector in ("", "-"):
        joined = connector.join(bits)
        aliases += [joined, joined[:-1]]
    return aliases


class FILESIZE(NUMERIC):
    """NUMERIC class with humanized filesize parser."""

    LOG = get_logger("FILESIZE")

    @staticmethod
    def _parse_size(value):
        """Parse the value for size suffixes."""
        try:
            value = str(parse_size(value))
        except InvalidSize as exc:
            FILESIZE.LOG.debug(exc)
        return value

    def parse_query(self, fieldname, qstring, boost=1.0):
        """Parse one term."""
        qstring = self._parse_size(qstring)
        return super().parse_query(fieldname, qstring, boost=boost)

    def parse_range(self, fieldname, start, end, startexcl, endexcl, boost=1.0):
        """Parse range terms."""
        if start:
            start = self._parse_size(start)
        if end:
            end = self._parse_size(end)
        return super().parse_range(
            fieldname, start, end, startexcl, endexcl, boost=boost
        )


class CodexSearchBackend(WhooshSearchBackend, WorkerBaseMixin):
    """Custom Whoosh Backend."""

    FIELDMAP = {
        "characters": ["category", "character"],
        "created_at": ["created"],
        "creators": ["author", "authors", "contributor", "contributors", "creator"],
        "community_rating": gen_multipart_field_aliases("community_rating"),
        "critical_rating": gen_multipart_field_aliases("critical_rating"),
        "genres": ["genre"],
        "locations": ["location"],
        "name": ["title"],
        "read_ltr": ["ltr"],
        "series_groups": gen_multipart_field_aliases("series_groups"),
        "scan_info": ["scan"],
        "story_arcs": gen_multipart_field_aliases("story_arcs"),
        "tags": ["tag"],
        "teams": ["team"],
        "updated_at": ["updated"],
    }
    RESERVED_CHARACTERS = ()
    RESERVED_WORDS = ()
    FIELD_ALIAS_PLUGIN = FieldAliasPlugin(FIELDMAP)
    OPERATORS_PLUGIN = OperatorsPlugin(
        ops=None,
        clean=False,
        And=r"(?i)(?<=\s)AND(?=\s)",
        Or=r"(?i)(?<=\s)OR(?=\s)",
        AndNot=r"(?i)(?<=\s)ANDNOT(?=\s)",
        AndMaybe=r"(?i)(?<=\s)ANDMAYBE(?=\s)",
        Not=r"(?i)(^|(?<=(\s|[()])))NOT(?=\s)",
        Require=r"(?i)(^|(?<=\s))REQUIRE(?=\s)",
    )
    STATUS_FINISH_TYPES = frozenset(
        (
            SearchIndexStatusTypes.SEARCH_INDEX_CLEAR,
            SearchIndexStatusTypes.SEARCH_INDEX_UPDATE,
        )
    )
    WRITERARGS = {
        "limitmb": 512,
        "procs": cpu_count(),
        # Bug in Whoosh means procs > 1 needs multisegment
        # https://github.com/mchaput/whoosh/issues/35
        "multisegement": True,
    }
    WRITER_PERIOD = 60 * 5
    WRITER_LIMIT = 10000
    COMMITARGS_MERGE_SMALL = {"merge": True}
    COMMITARGS_NO_MERGE = {"merge": False}

    def __init__(self, connection_alias, **connection_options):
        """Init worker queues."""
        super().__init__(connection_alias, **connection_options)
        self.log = getLogger(self.__class__.__name__)
        self.log.propagate = False

    def build_schema(self, fields):
        """Customize schema fields."""
        content_field_name, schema = super().build_schema(fields)

        # Add accent leniency to all text field.
        for _, field in schema.items():
            if isinstance(field, TEXT):
                field.analyzer = (
                    StandardAnalyzer()
                    | CharsetFilter(accent_map)
                    | StemFilter(cachesize=-1)
                )

        # Replace size field with FILESIZE type.
        old_field = schema["size"]
        schema.remove("size")
        schema.add(
            "size",
            FILESIZE(
                stored=old_field.stored,
                numtype=old_field.numtype,
                field_boost=old_field.format.field_boost,
            ),
        )

        return (content_field_name, schema)

    def setup(self, dateparser=True):
        """Add extra plugins."""
        super().setup()
        self.parser.replace_plugin(self.OPERATORS_PLUGIN)
        plugins = [self.FIELD_ALIAS_PLUGIN, GtLtPlugin]
        if dateparser:
            # the dateparser plugin won't pickle for
            # multiprocessing
            plugins += [DateParserPlugin(basedate=now())]
        self.parser.add_plugins(plugins)

    def get_writer(self, commitargs=COMMITARGS_NO_MERGE):
        """Get a writer."""
        self.index = self.index.refresh()
        writer = CodexWriter(
            self.index,
            limit=self.WRITER_LIMIT,
            period=self.WRITER_PERIOD,
            writerargs=self.WRITERARGS,
            commitargs=commitargs,
        )
        return writer

    def update(self, index, iterable, commit=True):
        """Update index, but with writer options."""
        num_objs = len(iterable)
        if not num_objs:
            self.log.debug("Search index nothing to update.")
            return 0

        if not self.setup_complete:
            self.setup(False)

        if not index:
            index = ComicIndex()

        # batch remove before update
        remove_pks = iterable.values_list("pk", flat=True)
        self.remove_batch_pks(remove_pks)

        writer = self.get_writer()

        for obj in iterable:
            try:
                doc = index.full_prepare(obj)
            except SkipDocument:
                self.log.debug("Indexing for object `%s` skipped", obj)
            else:
                # Really make sure it's unicode, because Whoosh won't have it any
                # other way.
                for key in doc:
                    doc[key] = self._from_python(doc[key])

                # Document boosts aren't supported in Whoosh 2.5.0+.
                if "boost" in doc:
                    del doc["boost"]

                try:
                    # add instead of update because of above batch remove
                    writer.add_document(**doc)
                except Exception as exc:
                    if not self.silently_fail:
                        raise

                    # We'll log the object identifier but won't include the actual
                    # object to avoid the possibility of that generating encoding
                    # errors while processing the log message:
                    self.log.warning(
                        f"Search index adding document {exc} pk:{obj.pk}",
                    )
        try:
            if num_objs:
                self.log.debug("Search index starting final commit.")
            else:
                self.log.debug("Search index update cancelling nothing to update.")
                writer.cancel()

            writer.close()
        except Exception as exc:
            self.log.warning(
                "Exception during search index writer final commit or cancel."
            )
            self.log.exception(exc)
        return num_objs

    def remove_batch_pks(self, pks, inverse=False, sc=None):
        """Remove a large batch of docs by pk from the index."""
        if not pks.count() and not inverse:
            return
        if not self.setup_complete:
            self.setup(False)

        writer = self.get_writer()
        try:
            pk_terms = [Term(DJANGO_ID, str(pk)) for pk in pks]
            query = Or(pk_terms)
            if inverse:
                query = Not(query)
            count = writer.delete_by_query(query, sc=sc)
        except Exception as exc:
            self.log.warning(
                f"Search index removing documents by query {inverse=} {exc}"
            )
            self.log.exception(exc)
            count = 0
        writer.close()
        return count

    def optimize(self):
        """Optimize the index."""
        if not self.setup_complete:
            self.setup(False)
        self.index = self.index.refresh()
        self.index.optimize(**self.WRITERARGS)

    def merge_small(self):
        """Merge small segments of the index."""
        if not self.setup_complete:
            self.setup(False)
        self.index = self.index.refresh()
        writer = self.get_writer({"merge": True})
        writer.close()
